import asyncio
import re
import unittest

from kobold import (
        assertions,
        compare,
        doubles,
        swap)

class TestException(Exception):
    pass

def test_function(*args, **kwargs):
    return 'test_function'

class Host(object):
    def __init__(self, *args, **kwargs):
        for (name, value) in kwargs.items():
            setattr(self, name, value)

    def test(self, *args, **kwargs):
        return 'original'

class TestRoutableStubFunction(unittest.TestCase):
    def setUp(self):
        self.safe_swap = swap.SafeSwap()

    def tearDown(self):
        self.safe_swap.rollback()

    def test_kwarg_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={'kwarg' : 1},
            stub_type='value',
            stub_value=2)
        stub_function.add_route(
            condition={'kwarg' : 3},
            stub_type='value',
            stub_value=4)

        self.assertEqual(2, stub_function(arg=1, kwarg=1))
        self.assertEqual(4, stub_function(arg=1, kwarg=3))

    def test_arg_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition=(1,),
            stub_type='value',
            stub_value=2)
        stub_function.add_route(
            condition=(3,),
            stub_type='value',
            stub_value=4)

        self.assertEqual(2, stub_function(1))
        self.assertEqual(4, stub_function(3))

    def test_self_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={
                'self': {'a': 1}},
            stub_type='value',
            stub_value=1)
        stub_function.add_route(
            condition={
                'self': {'a': 2}},
            stub_type='value',
            stub_value=2)
        self.safe_swap.swap(
            Host,
            'test',
            stub_function,
            default_original=True)

        host1 = Host(a=1)
        host2 = Host(a=2)
        self.assertEqual(
                1,
                host1.test())
        self.assertEqual(
                2,
                host2.test())

    def test_arg_and_kwarg_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={'args' : (1,),
                       'kwargs' : {'kwarg' : 1}},
            stub_type='value',
            stub_value=2)
        stub_function.add_route(
            condition={'args' : (3,),
                       'kwargs' : {'kwarg' : 3}},
            stub_type='value',
            stub_value=4)

        self.assertEqual(2, stub_function(1, kwarg=1))
        self.assertEqual(4, stub_function(3, kwarg=3))



    def test_kwarg_with_pattern(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={'kwarg' : re.compile('.*pattern1.*')},
            stub_type='value',
            stub_value=1)

        self.assertEqual(1, stub_function('_', kwarg='something_with_pattern1'))
        self.assertRaises(
                doubles.StubRoutingException,
                stub_function,
                arg='_',
                kwarg='something_without_the_pattern')

    def test_arg_with_pattern(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition=(re.compile('.*pattern1.*'),),
            stub_type='value',
            stub_value=1)

        self.assertEqual(1, stub_function('something_with_pattern1', kwarg='_'))
        self.assertRaises(
            doubles.StubRoutingException,
            stub_function,
            'something_without_the_pattern',
            kwarg='_')

    def test_arg_is_dict(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition=({'a': 1},),
            stub_type='value',
            stub_value=1)

        self.assertEqual(1, stub_function({'a': 1}))

    def test_rotate(self):
        stub_function = doubles.RoutableStubFunction()
        route = doubles.RotatingRoute(
                    condition=('a',),
                    routes=[
                     ('value', 1),
                     ('value', 2)])
                     
        stub_function.add_route(route=route)

        self.assertEqual(1, stub_function('a'))
        self.assertEqual(2, stub_function('a'))


class TestStubFunction(unittest.TestCase):
    def test_raises(self):
        e = TestException()
        stub_function = doubles.StubFunction(raises=e)
        self.assertRaises(
            TestException,
            stub_function)

    def test_calls(self):
        stub_function = doubles.StubFunction(calls=test_function)
        self.assertEqual('test_function', stub_function())

    def test_returns(self):
        stub_function = doubles.StubFunction(returns=1)
        self.assertEqual(1, stub_function())

class TestSpyFunction(unittest.TestCase):
    def test_reset(self):
        spy_function = doubles.SpyFunction(returns=1)
        self.assertEqual(1, spy_function(1))
        self.assertEqual(1, spy_function("apple"))
        self.assertEqual(1, spy_function(keyword="orange"))
        spy_function.reset()

        self.assertEqual([],
                          spy_function.calls)

    def test_spy(self):
        spy_function = doubles.SpyFunction(returns=1)
        self.assertEqual(1, spy_function(1))
        self.assertEqual(1, spy_function("apple"))
        self.assertEqual(1, spy_function(keyword="orange"))

        self.assertEqual(
            [((1,), {}, 1),
             (("apple",), {}, 1),
             ((), {'keyword' : 'orange'}, 1)],
            [x.as_tuple() for x in spy_function.calls])

class TestSpyCall(unittest.TestCase):
    def test_input(self):
        spy_function = doubles.SpyFunction(returns=1)
        spy_function('arg1', kwarg1='kwarg1_value')
        assertions.assert_match(
            (('arg1',), {'kwarg1': 'kwarg1_value'}),
            spy_function.calls[0].input())

    def test_as_tuple_return(self):
        spy_function = doubles.SpyFunction(returns=1)
        spy_function('arg1', kwarg1='kwarg1_value')
        assertions.assert_match(
            (('arg1',), {'kwarg1': 'kwarg1_value'}, 1),
            spy_function.calls[0].as_tuple())

    def test_as_tuple_raise(self):
        exception = LookupError()
        spy_function = doubles.SpyFunction(raises=exception)
        with self.assertRaises(LookupError):
            spy_function('arg1', kwarg1='kwarg1_value')
        assertions.assert_match(
            (('arg1',),
             {'kwarg1': 'kwarg1_value'},
             compare.NotPresent,
             exception),
            spy_function.calls[0].as_tuple())

    def test_as_dict_return(self):
        spy_function = doubles.SpyFunction(returns=1)
        spy_function('arg1', kwarg1='kwarg1_value')
        assertions.assert_match(
            {'args': ('arg1',),
             'kwargs': {'kwarg1': 'kwarg1_value'},
             'returned': 1,
             'raised': compare.NotPresent,
             'from_coroutine': False},
            spy_function.calls[0].as_dict())

    def test_as_dict_raise(self):
        exception = LookupError()
        spy_function = doubles.SpyFunction(raises=exception)
        with self.assertRaises(LookupError):
            spy_function('arg1', kwarg1='kwarg1_value')
        assertions.assert_match(
            {'args': ('arg1',),
             'kwargs': {'kwarg1': 'kwarg1_value'},
             'returned': compare.NotPresent,
             'raised': exception,
             'from_coroutine': False},
            spy_function.calls[0].as_dict())

    def test_as_dict_coroutine(self):
        spy_function = doubles.SpyCoroutine(returns=1)
        asyncio.get_event_loop().run_until_complete(
            spy_function('arg1', kwarg1='kwarg1_value'))
        assertions.assert_match(
            {'args': ('arg1',),
             'kwargs': {'kwarg1': 'kwarg1_value'},
             'returned': 1,
             'raised': compare.NotPresent,
             'from_coroutine': True},
            spy_function.calls[0].as_dict())


class TestRoutableSpyFunction(unittest.TestCase):
    def test_delegation(self):
        spy_function = doubles.SpyFunction(
                stub_function_factory=doubles.RoutableStubFunction)
        spy_function.add_route(
                condition=(1, 1),
                stub_type='value',
                stub_value=1)
        self.assertEqual(1, spy_function(1, 1))
        self.assertRaises(doubles.StubRoutingException,
                          spy_function,
                          1,
                          2)
        assertions.assert_match(
            [((1, 1), 
              {},
              1),
             ((1, 2), 
              {},
              compare.NotPresent,
              compare.DontCare(
                  rule='isinstance',
                  of_class=doubles.StubRoutingException))],
            [x.as_tuple() for x in spy_function.calls])



import collections
import unittest
from kobold import assertions

class TestAssertEqual(unittest.TestCase):
    def test_empty_hashes(self):
        assertions.assert_equal({}, {})

    def test_distinct_keys(self):
        self.assertRaises(
                AssertionError,
                assertions.assert_equal,
                {'a' : 1},
                {'b' : 2})
        
Response = collections.namedtuple('Response', 'headers status_code data')

class TestAssertResponseMatches(unittest.TestCase):
    def test_empty_body(self):
        actual = Response(headers={}, status_code=200, data={})
        assertions.assert_response_matches({'body' : {},
                                            'status_code' : 200,
                                            'headers' : {}}, actual)

    def test_omit_status_and_headers(self):
        actual = Response(headers={}, status_code=200, data={})
        assertions.assert_response_matches({'body' : {}}, actual)

    def test_equal_bodies(self):
        actual = Response(
                headers={}, 
                status_code=200, 
                data={'key' : 'value'})
        assertions.assert_response_matches({'body' : {'key' : 'value'},
                                            'status_code' : 200,
                                            'headers' : {}}, actual)

    def test_unequal_bodies(self):
        actual = Response(
                headers={}, 
                status_code=200, 
                data={'key' : 'value'})
        self.assertRaises(
                AssertionError,
                assertions.assert_response_matches,
                {'body' : {'key' : 'anothervalue'},
                           'status_code' : 200,
                           'headers' : {}}, 
                actual)

    def test_unequal_headers(self):
        actual = Response(
                headers={'header' : 'value'}, 
                status_code=200, 
                data={'key' : 'value'})
        self.assertRaises(
                AssertionError,
                assertions.assert_response_matches,
                {'body' : {'key' : 'value'},
                           'status_code' : 200,
                           'headers' : {'header' : 'anothervalue'}}, 
                actual)




import unittest

import kobold
from kobold import compare

class TestMultiMatch(unittest.TestCase):
    def test_ordered_basic(self):
        # We can use a MultiMatch to match against multiple
        # elements of a list
        actual = [
            {'color': 'blue',
             'shape': 'square'},
            {'color': 'blue',
             'shape': 'triangle'},
            {'color': 'blue',
             'shape': 'circle'}]
        diff = compare.compare(
            [compare.MultiMatch(
                {'color': 'blue'})],
            actual,
            type_compare={
                'hash': 'existing',
                'ordered': True})
        kobold.assertions.assert_match(
            'match',
            diff)

    def test_ordered_normal_element_and_multi_match(self):
        actual = [
            {'color': 'blue',
             'shape': 'square'},
            {'color': 'blue',
             'shape': 'triangle'},
            {'color': 'blue',
             'shape': 'circle'}]
        # Don't let a MultiMatch prevent a non-MultiMatch element
        # from being matched, simply because it's first in the list
        diff = compare.compare(
            [compare.MultiMatch(
                {'color': 'blue'}),
             {'shape': 'circle'}],
            actual,
            type_compare={
                'hash': 'existing',
                'ordered': True})
        kobold.assertions.assert_match(
            'match',
            diff)

    def test_ordered_two_multimatches(self):
        actual = [
            {'color': 'blue',
             'shape': 'square'},
            {'color': 'blue',
             'shape': 'triangle'},
            {'color': 'blue',
             'shape': 'circle'}]
        # Don't let a MultiMatch consume all the matches from another
        # MultiMatch, simply becuase it's first in the list
        diff = compare.compare(
            [compare.MultiMatch(
                {'color': 'blue'}),
             compare.MultiMatch(
                {'shape': 'circle'})],
            actual,
            type_compare={
                'hash': 'existing',
                'ordered': True})
        kobold.assertions.assert_match(
            'match',
            diff)

    # Unordered
    def test_unordered_basic(self):
        # We can use a MultiMatch to match against multiple
        # elements of a list
        actual = [
            {'color': 'blue',
             'shape': 'square'},
            {'color': 'blue',
             'shape': 'triangle'},
            {'color': 'blue',
             'shape': 'circle'}]
        diff = compare.compare(
            [compare.MultiMatch(
                {'color': 'blue'})],
            actual,
            type_compare={
                'hash': 'existing',
                'ordered': False})
        kobold.assertions.assert_match(
            'match',
            diff)

    def test_unordered_normal_element_and_multi_match(self):
        actual = [
            {'color': 'blue',
             'shape': 'square'},
            {'color': 'blue',
             'shape': 'triangle'},
            {'color': 'blue',
             'shape': 'circle'}]
        # Don't let a MultiMatch prevent a non-MultiMatch element
        # from being matched, simply because it's first in the list
        diff = compare.compare(
            [compare.MultiMatch(
                {'color': 'blue'}),
             {'shape': 'circle'}],
            actual,
            type_compare={
                'hash': 'existing',
                'ordered': False})
        kobold.assertions.assert_match(
            'match',
            diff)

    def test_unordered_two_matches(self):
        actual = [
            {'color': 'blue',
             'shape': 'square'},
            {'color': 'blue',
             'shape': 'triangle'},
            {'color': 'blue',
             'shape': 'circle'}]
        # Don't let a MultiMatch consume all the matches from another
        # MultiMatch, simply becuase it's first in the list
        diff = compare.compare(
            [compare.MultiMatch(
                {'color': 'blue'}),
             compare.MultiMatch(
                {'shape': 'circle'})],
            actual,
            type_compare={
                'hash': 'existing',
                'ordered': False})
        kobold.assertions.assert_match(
            'match',
            diff)



import json
import re

from kobold import compare


def parse_body(content_type, content):
    if content_type is None:
        return content
    if content_type == 'application/json':
        return json.loads(content)
    elif re.compile('application/json;.*').search(content_type):
        return json.loads(content)
    else:
        return content


def response_matches(expected,
                     response,
                     type_compare=None):
    '''
    Compare two HTTP responses (using kobold.compare).
    A type_compare may be provided, but if none is set,
    the default is for hashes to be compared using "full" mode,
    and for lists to be compared using "ordered" mode.
    Headers, however, are supplied an override to ensure that 
    they are compared in "existing" mode.

    The first two arguments are the expected and actual response.
    Responses are expected to be a hash, with three keys: status_code,
    body and headers.  By default, the status_code of the response is 
    expected to be 200 (ie. success of the request is enforced), and no 
    headers are enforced.  

    Body will be a string, unless the actual response has a 
    Content-Type header of "application/json".  In that case,
    we parse the JSON and set body to be the parsed data structure.
    Remember, by default all keys in the body will be compared.
    type_compare can be set to "existing" to change this behavior.

    Ultimately, the expected and response hashes are compared using
    kobold.compare, and the result is returned.  In the case of a match,
    the result will be the string "match".  In the case of a mismatch,
    the result will be a tuple of two elements - the first describing
    the mismatched values in the first argument, and the second
    describing the mismatched values in the second argument.
    '''

    if type_compare is None:
        type_compare = {
                'hash' : 'full',
                'ordered' : True}
    else:
        if isinstance(type_compare, str):
            type_compare = {
                    'hash' : type_compare,
                    'ordered' : True}

    default_expected =\
        {'status_code' : 200,
         'headers' : {}}
    default_expected_headers =\
        {'__compare' : 'existing'}

    for key, value in default_expected.items():
        expected.setdefault(key, value)

    for key, value in default_expected_headers.items():
        expected['headers'].setdefault(key, value)

    content_type = response.headers.get('Content-Type')
    actual = {'status_code' : response.status_code,
              'headers' : response.headers,
              'body' : parse_body(content_type, response.data)}

    return compare.compare(expected, 
                   actual,
                   type_compare)



import base64
import json
import pickle

import kobold


class ParsingHint(object):
    '''Tells the comparison function to parse the 
       second argument ("actual") in a specific way to
       get a simple data-structure'''


    def __init__(self, payload):
        self.payload = payload

    def __repr__(self):
        return '{}(payload={})'.format(
            self.__class__.__name__,
            self.payload)

    def parse(self, thing_to_parse):
        if thing_to_parse is kobold.NotPresent:
            return kobold.NotPresent

        result = self.sub_parse(thing_to_parse)
        return result

    def sub_parse(self, thing_to_parse):
        return thing_to_parse


class TypeCompareHint(ParsingHint):
    def __init__(self, payload, type_compare):
        self.payload = payload
        self.type_compare = type_compare

    def parse(self, *args, **kwargs):
        raise NotImplementedError(
            'TypeCompareHint should never be used in comparisons!')


class JSONParsingHint(ParsingHint):
    def sub_parse(self, thing_to_parse):
        try:
            return json.loads(thing_to_parse)
        except (TypeError, json.decoder.JSONDecodeError):
            raise kobold.InvalidMatch


class ObjectDictParsingHint(ParsingHint):
    def sub_parse(self, thing_to_parse):
        return thing_to_parse.__dict__


class ObjectAttrParsingHint(ParsingHint):
    def sub_parse(self, thing_to_parse):
        attr_dict = {}
        for (key, value) in self.payload.items():
            attr_dict[key] = getattr(
                    thing_to_parse, 
                    key, 
                    kobold.NotPresent)
        return attr_dict


class Base64Hint(ParsingHint):
    def sub_parse(self, thing_to_parse):
        return base64.b64decode(thing_to_parse)

class PickleParsingHint(ParsingHint):
    def sub_parse(self, thing_to_parse):
        return pickle.loads(thing_to_parse)

class MultiMatch(ParsingHint):
    def __init__(self, payload):
        self.payload = payload
        self.count = 0

    def add_match(self, element):
        self.count += 1

    def matched(self):
        return self.count > 0





import asyncio
import inspect

from kobold import compare, doubles


class SafeSwap(object):
    '''Instances of this class manage runtime replacements of members,
       and enable all replacements to be undone with the rollback method'''
       
    def __init__(self):
        self.registry = {}

    def swap(self, 
             host,
             member_name,
             new_member,
             default_original=False,
             pass_original=False,
             swap_type='member'):
        '''Given a host object, a member name, and some other object,
           replace the member of that name on that host with the given object.
           This can be undone with rollback()'''
        key = self.get_key(host, member_name)
        if not key in self.registry:
            if swap_type == 'member':
                original = getattr(host, member_name, compare.NotPresent)
            elif swap_type == 'key':
                original = host.get(member_name, compare.NotPresent)

            self.registry[key] = (
                host,
                original,
                swap_type)

        if pass_original:
            def stub_wrapper(*args, **kwargs):
                return new_member(
                    *args,
                    original_reference=self.registry[key][1],
                    **kwargs)
            swap_target = stub_wrapper
        else:
            swap_target = new_member
        
        if swap_type == 'member':
            if new_member is compare.NotPresent:
                delattr(host, member_name)
            else:
                setattr(host, member_name, swap_target)
        elif swap_type == 'key':
            if new_member is compare.NotPresent:
                if host.get(member_name, compare.NotPresent) is not compare.NotPresent:
                    del host[member_name]
            else:
                host[member_name] = swap_target

        if getattr(new_member, 'set_original_reference', None) is not None:
            new_member.set_original_reference(self.registry[key][1])

        if default_original:
            new_member.default_original(self.registry[key][1])

    def install_proxy(self,
                      host,
                      member_name,
                      proxy_factory=compare.NotPresent,
                      stub_function_factory=compare.NotPresent):
        '''Replace a member with a function that does something else,
           and then calls that original member (this is very similar
           to the idea of dynamically decorating a function at runtime, 
           except that the decoration will be undone when rollback is called.

           Typically, the idea here is to inject a spy in front of the
           original functionality.  So, the function will have the same
           result as it would without this replacement, except the spy remembers
           each call.'''
        key = self.get_key(host, member_name)
        registry_entry = self.registry.get(
            key,
            None)

        if registry_entry is None:
            original_member = getattr(host, member_name)
        else:
            original_member = registry_entry[1]

        if inspect.iscoroutinefunction(original_member):
            if proxy_factory is compare.NotPresent:
                proxy_factory = doubles.SpyCoroutine
            if stub_function_factory is compare.NotPresent:
                stub_function_factory = doubles.StubCoroutine
        elif callable(original_member):
            if proxy_factory is compare.NotPresent:
                proxy_factory = doubles.SpyFunction
            if stub_function_factory is compare.NotPresent:
                stub_function_factory = doubles.StubFunction
        else:
            raise NotImplementedError(
                'Original member {}.{} is neither '
                'a coroutine nor a function'.format(
                    host,
                    member_name))

        proxy = proxy_factory(stub_function_factory=stub_function_factory)
        self.swap(host, member_name, proxy)
        (host, original_member, swap_type) = self.registry[key]
        proxy.stub_function.calls(original_member)
        return proxy

    def make_decorator(
            self,
            decorated_function,
            before=None,
            after=None,
            on_failure=None):
        if asyncio.iscoroutinefunction(decorated_function):
            async def decorator(*args, **kwargs):
                if before:
                    if asyncio.iscoroutinefunction(before):
                        await before(*args, **kwargs)
                    else:
                        before(*args, **kwargs)
                try:
                    ret = await decorated_function(*args, **kwargs)
                except Exception as e:
                    if on_failure:
                        return on_failure(e, *args, **kwargs)
                    else:
                        raise
                if after:
                    if asyncio.iscoroutinefunction(after):
                        ret = await after(ret, *args, **kwargs)
                    else:
                        ret = after(ret, *args, **kwargs)
                return ret
        else:
            def decorator(*args, **kwargs):
                if before:
                    before(*args, **kwargs)
                try:
                    ret = decorated_function(*args, **kwargs)
                except Exception as e:
                    if on_failure:
                        return on_failure(e, *args, **kwargs)
                    else:
                        raise
                if after:
                    ret = after(ret, *args, **kwargs)
                return ret

        return decorator

    def install_decorator(self,
                          host,
                          member_name,
                          before=None,
                          after=None,
                          on_failure=None):
        decorator = self.make_decorator(
            getattr(host, member_name),
            before=before,
            after=after,
            on_failure=on_failure)
        self.swap(host, member_name, decorator)
        return decorator


    def unswap(self, host, member_name):
        '''Rollback a specific replacement'''
        key = self.get_key(host, member_name)
        (host, original_member, swap_type) = self.registry[key]
        self._unswap(host, member_name, original_member, swap_type)
        del self.registry[key]

    def _unswap(self, host, member_name, original_member, swap_type):
        if swap_type == 'member':
            if original_member is compare.NotPresent:
                delattr(host, member_name)
            else:
                setattr(host, member_name, original_member)
        elif swap_type == 'key':
            if original_member is compare.NotPresent:
                if (host.get(member_name, compare.NotPresent)
                        is not compare.NotPresent):
                    del host[member_name]
            else:
                host[member_name] = original_member

    def rollback(self):
        '''Rollback all replacements (at the end of a test, for instance)'''
        for ((host_name, member_name),
             (host, original_member, swap_type)) in self.registry.items():
            self._unswap(host, member_name, original_member, swap_type)
        self.registry = {}

    def get_key(self,
                host,
                member_name):
        return (id(host), member_name)
